/*
 * Copyright (c) 2003 Marcus Glocker
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "../libmy/file.h"
#include "../libmy/str.h"

#include "config.h"
#include "proto.h"
#include "extern.h"

int		sfd;
struct cfg	config;

int
conreg(void)
{
	if ((sfd = net_connect(config.ircserver, config.ircserverport)) == -1)
		return -1;

	irc_regpost(sfd);

	return 0;
}

void
signal_handler(const int signal)
{
	switch(signal) {
		case SIGCHLD: {
			int	status;
			
			while (waitpid(-1, &status, WNOHANG) > 0) {
			}

			break;
		}
		case SIGTERM:
			net_snd(sfd, "QUIT :Process terminated !\n");
			flogd(config.filelog, "*** bot stopped (pid %d)\n",
			    ppid);
			close(sfd);
			exit(0);
	}
}

void
usage(int mode)
{
	extern char	*__progname;

	if (mode != 1) {
		fprintf(stderr, "usage: %s ", __progname);
		fprintf(stderr, "[-v] [-c config file]\n");
	}

	if (mode == 1)
		fprintf(stderr, "bot 1.6\n");

	exit(1);
}

int
main(int argc, char *argv[])
{
	int		i, j, k, ch, fd, r = 0;
	char		in[LINE_NR*LINE_SIZE+1], tmp[LINE_SIZE+1];
	char		lines[LINE_NR][LINE_SIZE+1];
	fd_set		readfds;
	struct timeval	tv;

	if (argc < 2)
		usage(0);

	while ((ch = getopt(argc, argv, "vc:")) != -1) {
		switch (ch) {
			case 'v':
				usage(1);
				exit(0);
			case 'c':
				strlcpy(config.file, optarg,
				    sizeof(config.file));
				break;
			default:
				usage(0);
				break;
		}
	}
	if (optind == 1)
		usage(0);

	/* check if configuration file exists */
	if ((fd = open(config.file, O_RDONLY)) == -1) {
		fprintf(stderr, "Can't open configuration file: %s !\n",
		    config.file);
		exit(1);
	}

	/* mandatory configuration parameters */
	if (fparse(tmp , "ircserverport", config.file,
	    sizeof(tmp)) == -1) {
		fprintf(stderr, "Missing <ircserverport> in config file.\n");
		r = -1;
	} else {
		config.ircserverport = atoi(tmp);
	}
	if (fparse(config.ircserver, "ircserver", config.file,
	    sizeof(config.ircserver)) == -1) {
		fprintf(stderr, "Missing <ircserver> in config file.\n");
		r = -1;
	}
	if (fparse(config.botnick, "botnick", config.file,
	    sizeof(config.botnick)) == -1) {
		fprintf(stderr, "Missing <botnick> in config file.\n");
		r = -1;
	}
	if (fparse(config.realname, "realname", config.file,
	    sizeof(config.realname)) == -1) {
		fprintf(stderr, "Missing <realname> in config file.\n");
		r = -1;
	}
	if (fparse(config.channel, "channel", config.file,
	    sizeof(config.channel)) == -1) {
		fprintf(stderr, "Missing <channel> config file.\n");
		r = -1;
	}
	if (fparse(config.botroot, "botroot", config.file,
	    sizeof(config.botroot)) == -1) {
		fprintf(stderr, "Missing <botroot> in config file.\n");
		r = -1;
	}
	if (fparse(config.filelog, "filelog", config.file,
	    sizeof(config.filelog)) == -1) {
		fprintf(stderr, "Missing <filelog> in config file.\n");
		r = -1;
	}
	if (fparse(config.fileuser, "fileuser", config.file,
	    sizeof(config.fileuser)) == -1) {
		fprintf(stderr, "Missing <fileuser> in config file.\n");
		r = -1;
	}
	if (fparse(config.admin, "admin", config.file,
	    sizeof(config.admin)) == -1) {
		fprintf(stderr, "Missing <admin> in config file.\n");
		r = -1;
	}

	/* optional configuration parameters */
	if (fparse(config.topic, "topic", config.file,
	    sizeof(config.file)) == -1)
		strlcpy(config.topic, "0", sizeof(config.topic));
	if (fparse(config.filedb, "filedb", config.file,
	    sizeof(config.filedb)) == -1)
		strlcpy(config.filedb, "0", sizeof(config.topic));
	if (fparse(config.moddir, "moddir", config.file,
	    sizeof(config.moddir)) == -1) {
		strlcpy(config.moddir, "0", sizeof(config.moddir));
	} else {
		snprintf(tmp, sizeof(config.moddir), "%s/%s",
		    config.botroot, config.moddir);
		snprintf(config.moddir, sizeof(config.moddir), "%s", tmp);
	}
	/* get local hostname */
	if (gethostname(config.hostname, sizeof(config.hostname)) == -1) {
		fprintf(stderr, "Can't get local hostname.\n");
		r = -1;
	}
	if (r == -1)
		exit(1);

	/* daemonize */
	if (sys_daemon() != 0)
		exit(1);

	/* signal handler */
	signal(SIGTERM, signal_handler);
	signal(SIGCHLD, signal_handler);

	/* load modtab */
	if (strcmp(config.moddir, "0") != 0)
		sys_modtab_load();

	/* connect irc server */
	if (conreg() == -1) {
		flogd(config.filelog, "*** Can't connect %s !\n",
		    config.ircserver);
		exit(1);
	}

	while (1) {
		/* initialize chars */
		memset(in, 0, sizeof(in));
		memset(tmp, 0, sizeof(tmp));
		for (i = 0; i < LINE_NR; i++)
			memset(lines[i], 0, sizeof(lines[i]));

		tv.tv_sec = 1;
		tv.tv_usec = 0;
		FD_ZERO(&readfds);
		FD_SET(sfd, &readfds);
		r = select(sfd + 1, &readfds, NULL, NULL, &tv);
		if (r == 0) {
			if (strcmp(config.moddir, "0") != 0)
				sys_modtab_check(sfd);
			continue;
		}
		if (r == -1)
			continue;

		/* receive data */
		r = recv(sfd, in, sizeof(in), 0);

		/* check connection to irc server closed */
		if (r == 0) {
			flogd(config.filelog, "*** Disconnected from %s !\n",
			    config.ircserver);
			flogd(config.filelog, "*** Try reconnecting ...\n");
			close(sfd);
			/* try reconnecting */
			while (conreg() == -1) {
				sleep(10);
				flogd(config.filelog, "*** Still trying ...\n");
			}
			/* reconnected */
			continue;
		}
		/* error on data receive */
		if (r == -1)
			continue;

		/* size check for lines */
		r = strcutl(tmp, in, 1, sizeof(tmp));
		if (r > LINE_NR) {
			flogd(config.filelog, "*** Too many lines received at "
			    "once, abort processing !\n");
			continue;
		}

		/* we need to cut the received into single lines */
		for (i = 0, j = 0, k = 0; in[i] != '\0'; i++) {
			/* size check for one line */
			if (j == LINE_SIZE) {
				flogd(config.filelog, "*** Line too large, "
				    "abort processing !\n");
				continue;
			}
			if (in[i] != '\n') {
				tmp[j] = in[i];
				j++;
			} else {
				tmp[j] = in[i];
				strlcpy(lines[k], tmp, sizeof(lines[k]));
				flogd(config.filelog, "%s", lines[k]);
				k++;
				j = 0;
				memset(tmp, 0, sizeof(tmp));
			}
		}

		/* and process every single line */
		for (i = 0; i < k; i++) {
			r = irc_listen(lines[i], sfd);

			/* server action */
			if (r == 0)
				r = 0;
			/* ignore myself */
			if (r == 1)
				r = 1;
			/* 001 */
			if (r == 2)
				irc_regpre(lines[i], sfd);
			/* join */
			if (r == 3)
				irc_join(lines[i], sfd);
			/* notice */
			if (r == 4)
				irc_notice(lines[i], sfd);
			/* part */
			if (r == 5)
				r = 5; /* not used yet */
			/* quit */
			if (r == 6)
				r = 6; /* not used yet */
			/* mode */
			if (r == 7)
				irc_mode(lines[i], sfd);
			/* kick */
			if (r == 8)
				irc_kick(lines[i], sfd);
			/* channel msg */
			if (r == 9)
				irc_privpub(lines[i], sfd);
			/* private msg */
			if (r == 10)
				irc_privprv(lines[i], sfd);
		}
	}

	return 0;
}
